// HPModel.cpp

#include "HPModel.h"
#include <map>
#include <set>


Point::Point(int x, int y, int z)
    :x(x),y(y),z(z) {}

Point::Point() {
    x = 0;
    y = 0;
    z = 0;
}


inline bool Point::operator==(const Point& other) const {
    return x == other.x && y == other.y && z == other.z;
}


inline bool Point::operator<(const Point& other) const {
    return (7 * x + 11 *  y + 13 * z) < (7 * other.x + 11 * other.y + 13 * other.z);
}

ScorePair::ScorePair(Point p1, Point p2)
    :p1(p1),p2(p2) {}

bool ScorePair::operator==(const ScorePair &other) const {
    return (p1 == other.p1 && p2 == other.p2) || (p1 == other.p2 && p2 == other.p1);
}

bool ScorePair::operator<(const ScorePair &other) const {
    int p1Score = p1.x * 7 + p1.y * 11 + p1.z * 13;
    int p2Score = p2.x * 7 + p2.y * 11 + p1.z * 13;

    int otherP1Score = other.p1.x * 7 + other.p1.y * 11 + other.p1.z * 13;
    int otherP2Score = other.p2.x * 7 + other.p2.y * 11 + other.p2.z * 13;

    return (p1Score + p2Score) < (otherP1Score + otherP2Score);
}


// Acid method definitions

Acid::Acid(const int x, const int y, const int z, const bool hydrophobic)
        : x(x), y(y), z(z), hydrophobic(hydrophobic) {
    point = Point(x,y,z);
}
Acid::Acid()
    : x(0), y(0), z(0), hydrophobic(false), point(0, 0, 0) {}

Point Acid::getPoint() {
    return point;
}

void Acid::print() const {
    std::cout << "Acid (" << x << ", " << y << ", " << z << ")\n";
    //std::cout << (hydrophobic ? " - Hydrophobic (H)" : " - Polar (P)") << "\n";
}

bool Acid::coords_match(Acid other) const {
    return other.x == x && other.y == y && other.z == z;
}

bool Acid::operator==(const Acid &other) const {
    return point == other.point;
}

bool Acid::operator<(const Acid &other) const {
    return point < other.point;
}

// HPModel method definitions

HPModel::HPModel(const std::string &input_sequence) : HPModel(input_sequence, {}) {}

HPModel::HPModel(const std::string &input_sequence, const std::string &folds) {
    currentDirection = NORTH;
    previousDirection = NORTH;
    rotation = NO_ROTATION;
    if (input_sequence.size() < 2) {
        throw std::invalid_argument("Sequence length must be at least 2");
    }
    for (char c: input_sequence) {
        if (c != 'h' && c != 'p') {
            throw std::invalid_argument("Invalid character found in sequence, '" + std::to_string(c) +
                                        "' is not a valid char, only 'h' and 'p' allowed");
        }
    }
    sequence = input_sequence;
    acidMap = std::map<Point, Acid>();
    scoringPairs = std::set<ScorePair>();
    if (!folds.empty()) {
        doFolds(folds);
    }
}

Acid HPModel::getHead() {
    if (acids.size() == 0) {
        throw std::out_of_range("Acid sequence empty, do not call getHead without elements\n");
    }
    return acids.back();
}

std::string HPModel::directionToString(DIRECTION direction) {
    switch (direction) {
        case FORWARD:
            return "FORWARD";
        case UP:
            return "UP";
        case DOWN:
            return "DOWN";
        case LEFT:
            return "LEFT";
        case RIGHT:
            return "RIGHT";
        default:
            return "UNKNOWN";
    }
}

bool HPModel::isInsertValid(Acid toInsert) {
    for (Acid p: acids) {
        if (toInsert.coords_match(p)) {
            return false;
        }
    }
    return true;
}

/**
 * Updates the rotation in 90 degree steps
 * POSITIVE = CLOCKWISE
 * NEGATIVE = COUNTER CLOCKWISE
 * @param amount amount of steps to rotate, for each step rotates 90 degrees, negative amount will rotate opposite direction
 */
void HPModel::rotate(int amount) {
    // Since a rotation step is 90 degrees, every 4 rotations will be 360 degrees
    int modAmount = amount % 4;
    int rotationTimes = modAmount;
    if (modAmount < 0) {
        rotationTimes = 4 - (-modAmount);
    }
    for (int i = 0; i < rotationTimes; i++) {
        switch (rotation) {
            case NO_ROTATION:
                rotation = ROTATE_90;
                break;
            case ROTATE_90:
                rotation = ROTATE_180;
                break;
            case ROTATE_180:
                rotation = ROTATE_270;
                break;
            case ROTATE_270:
                rotation = NO_ROTATION;
                break;
        }
    }
}

ABSOLUTE_DIRECTION
HPModel::getPlaneDirectionChange(ABSOLUTE_DIRECTION absolute_direction, DIRECTION relative_direction) {
    if (absolute_direction == ABS_UP || absolute_direction == ABS_DOWN || relative_direction == UP ||
        relative_direction == DOWN) {
        throw std::invalid_argument("Not a planar direction, only N/S/E/W allowed\n");
    }
    switch (relative_direction) {
        case FORWARD:
            return absolute_direction;
        case RIGHT:
            switch (absolute_direction) {
                case NORTH:
                    return EAST;
                case SOUTH:
                    return WEST;
                case EAST:
                    return SOUTH;
                case WEST:
                    return NORTH;
                default:
                    throw std::invalid_argument("Invalid absolute_direction, case RIGHT \n");
            }
        case LEFT:
            switch (absolute_direction) {
                case NORTH:
                    return WEST;
                case SOUTH:
                    return EAST;
                case EAST:
                    return NORTH;
                case WEST:
                    return SOUTH;
                default:
                    throw std::invalid_argument("Invalid absolute_direction, case LEFT\n");
            }
        default:
            throw std::invalid_argument("Invalid relative_direction in method getPlaneDirectionChange");
    }
}

/*
 * Has the side effect of updating previousDirection, because this update is non-trivial
 * Similarly updates the rotation, because up/down are the only folds that will do so
 */
ABSOLUTE_DIRECTION HPModel::getUpDownDirectionAndUpdatePrevious(DIRECTION relative_direction) {
    if (relative_direction != UP && relative_direction != DOWN) {
        throw std::invalid_argument("Invalid direction passed, must be either UP/DOWN\n");
    }
    ABSOLUTE_DIRECTION new_direction;
    switch (currentDirection) {
        case NORTH:
        case EAST:
        case SOUTH:
        case WEST:
            if (relative_direction == UP) {
                // NORTH/EAST/SOUTH/WEST -> UP
                new_direction = ABS_UP;
            } else {
                // NORTH/EAST/SOUTH/WEST -> DOWN
                new_direction = ABS_DOWN;
            }
            // Note: in this case we DO NOT want to update the previous direction
            break;
        case ABS_UP:
            switch (previousDirection) {
                case NORTH:
                case EAST:
                case SOUTH:
                case WEST:
                    if (relative_direction == UP) {
                        // UP -> UP
                        new_direction = invertDirection(previousDirection);
                        // After UP -> UP we will be "upside-down" and will need to rotate 180 degrees
                        rotate(2);
                    } else {
                        // UP -> DOWN
                        new_direction = previousDirection;
                        // No rotation required because this is equivalent to simply FORWARD
                    }
                    break;
                case ABS_UP:
                case ABS_DOWN:
                    throw std::invalid_argument("Both current and previous direction is UP/DOWN, illegal state\n");
            }
            break;
        case ABS_DOWN:
            switch (previousDirection) {
                case NORTH:
                case EAST:
                case SOUTH:
                case WEST:
                    if (relative_direction == UP) {
                        // DOWN -> UP
                        new_direction = previousDirection;
                        // No rotation required
                    } else {
                        // DOWN -> DOWN
                        new_direction = invertDirection(previousDirection);
                        // DOWN -> DOWN results in "upside-down" state just as UP -> UP does
                        rotate(-2);
                    }
                    break;
                case ABS_UP:
                case ABS_DOWN:
                    throw std::invalid_argument("Both current and previous direction is UP/DOWN, illegal state\n");
            }
            break;
    }
    return new_direction;
}

void HPModel::updateCurrentDirection(DIRECTION direction) {
    ABSOLUTE_DIRECTION new_direction;
    DIRECTION rotated_direction = getRotatedRelativeDirection(direction);
    switch (rotated_direction) {
        case FORWARD:
            new_direction = currentDirection;
            break;
        case RIGHT:
        case LEFT:
            switch (currentDirection) {
                case NORTH:
                case SOUTH:
                case EAST:
                case WEST:
                    new_direction = getPlaneDirectionChange(currentDirection, rotated_direction);
                    break;
                case ABS_UP:
                case ABS_DOWN:
                    // In this case we need to handle rotation changes, because moving "out" of an UP/DOWN forces a rotation
                    if (currentDirection == ABS_UP) {
                        // UP
                        if (rotated_direction == RIGHT) {
                            // UP -> RIGHT
                            rotate(1);
                        } else {
                            // UP -> LEFT
                            rotate(-1);
                        }
                    } else {
                        // DOWN
                        if (rotated_direction == RIGHT) {
                            // DOWN -> RIGHT
                            rotate(-1);
                        } else {
                            // DOWN -> LEFT
                            rotate(1);
                        }
                    }
                    new_direction = getPlaneDirectionChange(previousDirection, rotated_direction);
                    break;
                default:
                    throw std::invalid_argument(
                            "Enum value: " + directionToString(rotated_direction) + " is invalid\n");
            }
            break;
        case UP:
            //new_direction = ABS_UP;
            new_direction = getUpDownDirectionAndUpdatePrevious(rotated_direction);
            break;
        case DOWN:
            //new_direction = ABS_DOWN;
            new_direction = getUpDownDirectionAndUpdatePrevious(rotated_direction);
            break;
        default:
            throw std::invalid_argument("Enum value: " + directionToString(rotated_direction) + " is invalid\n");
    }

    // ABS_UP/ABS_DOWN are handled directly in getUpDownDirectionAndUpdatePrevious
    if ((new_direction == ABS_UP || new_direction == ABS_DOWN) &&
        (currentDirection == ABS_UP || currentDirection == ABS_DOWN)) {
        currentDirection = new_direction;
        return;
    }
    if ((new_direction == ABS_UP || new_direction == ABS_DOWN) &&
        (previousDirection == ABS_UP || previousDirection == ABS_DOWN)) {
        previousDirection = currentDirection;
        currentDirection = new_direction;
        return;
    }
    if (new_direction != previousDirection) {
        previousDirection = currentDirection;
    }
    currentDirection = new_direction;
}

bool HPModel::isCurrentAcidHydrophobic() {
    int size = acids.size();
    char stringElement = sequence[size];
    return stringElement == 'h';
}

void HPModel::fold(DIRECTION direction) {
    // An initial fold will *ALWAYS* require an extra point inserted (the 0th element of the sequence),
    // and this direction will always be 'forward'
    if (acids.empty()) {
        addPoint(FORWARD, isCurrentAcidHydrophobic());
    }
    int no_of_points = acids.size();

    // We allow n-1 folds, where n is the size of the sequence, assume we are inserting fold k (for k > 0)
    // At this point, there exists k folds, and no_of_points = k + 1 => k = no_of_points - 1
    // We want k <= n - 1, thus we need, no_of_points - 1 < n - 1 => no_of_points < n
    if (no_of_points >= sequence.size()) {
        throw std::invalid_argument(
                "Too many folds when trying to insert fold number " + std::to_string(no_of_points - 1));
    }
    addPoint(direction, isCurrentAcidHydrophobic());
}

std::vector<Acid> HPModel::getNearbyPotentialAcids(Acid currentAcid, Acid connectingAcid) {
    std::vector<Acid> res = std::vector<Acid>();
    Point currentPoint = currentAcid.getPoint();
    Point connectingPoint = connectingAcid.getPoint();

    // Check all non-diagonal directions except for the connecting point
    std::vector<Point> allNonDiagonal = std::vector<Point>();
        int x = currentPoint.x;
    int y = currentPoint.y;
    int z = currentPoint.z;
    // Note that we use relative directions here assuming "forward" is north.
    // This trivially reduces to any other rotation/facing direction.
    // Up
    Point up {x, y + 1, z};
    allNonDiagonal.emplace_back(up);
    // Down
    Point down {x, y - 1, z};
    allNonDiagonal.emplace_back(down);
    // Left
    Point left {x - 1, y, z};
    allNonDiagonal.emplace_back(left);
    // Right
    Point right {x + 1, y, z};
    allNonDiagonal.emplace_back(right);
    // Forward
    Point forward {x, y, z + 1};
    allNonDiagonal.emplace_back(forward);
    // Backwards
    Point backwards{x, y, z - 1};
    allNonDiagonal.emplace_back(backwards);

    for (auto p : allNonDiagonal) {
        // We skip the connecting point, as this can never score
        if (p == connectingPoint) {
            continue;
        }
        auto it = acidMap.find(p);
        if (it == acidMap.end()) {
            continue;
        }
        // First=key second=value
        Acid acid = it->second;
        // Important to emplace here so we do not mess with our acid map!
        res.emplace_back(acid);
    }
    return res;
}

bool existsInTupleMap(std::map<Acid, Acid> tupleMap, Acid acid1, Acid acid2) {
    return true;
}

void HPModel::updateScore(Acid latestAcid) {
    int acidLength = (int) acids.size(); // ULL -> int cast should be safe, if it is not, we have bigger problems
    // Absolute minimum length that can score is 4 acids (3 folds)
    if (acidLength < 4) {
        return;
    }
    // Adding a new acid can only ever score points if it is hydrophobic
    if (!latestAcid.hydrophobic) {
        return;
    }
    int connectingAcidIndex = acidLength - 2;
    Acid connectingAcid = acids[connectingAcidIndex];
    auto potentialScoringAcids = getNearbyPotentialAcids(latestAcid, connectingAcid);
    for (auto acid : potentialScoringAcids) {
        if (!acid.hydrophobic) {
            continue;
        }
        ScorePair pair = ScorePair(acid.getPoint(), latestAcid.getPoint());
        scoringPairs.insert(pair);
    }
}

void HPModel::insert(Acid acid) {
    acids.emplace_back(acid);
    acidMap.insert({acid.getPoint(), acid});
    updateScore(acid);
}

void HPModel::addPoint(DIRECTION direction, bool isHydrophobic) {
    if (acids.empty()) {
        Acid acid = Acid(0, 0, 0, isHydrophobic);
        insert(acid);
        updateCurrentDirection(direction);
        return;
    }

    Acid head = getHead();
    int newX = head.x, newY = head.y, newZ = head.z;
    updateCurrentDirection(direction);

    switch (currentDirection) {
        case NORTH:
            newZ++;
            break;
        case SOUTH:
            newZ--;
            break;
        case EAST:
            newX++;
            break;
        case WEST:
            newX--;
            break;
        case ABS_UP:
            newY++;
            break;
        case ABS_DOWN:
            newY--;
            break;
    }

    Acid toInsert = Acid(newX, newY, newZ, isHydrophobic);
    if (!isInsertValid(toInsert)) {
        throw std::invalid_argument("Unable to insert point in sequence\n");
    }
    insert(toInsert);
}

void HPModel::print() const {
    for (const Acid &p: acids) {
        p.print();
    }
}

std::vector<Acid> HPModel::getPoints() {
    return acids;
}

ABSOLUTE_DIRECTION HPModel::invertDirection(ABSOLUTE_DIRECTION direction) {
    switch (direction) {

        case NORTH:
            return SOUTH;
        case EAST:
            return WEST;
        case SOUTH:
            return NORTH;
        case WEST:
            return EAST;
        case ABS_UP:
        case ABS_DOWN:
            throw std::invalid_argument("Unable to invert direction of ABS_UP/ABS_DOWN directions\n");
            break;
    }
}

DIRECTION rotateRelativeDirection90Degrees(DIRECTION relativeDirection) {
    switch (relativeDirection) {
        case FORWARD:
            return FORWARD;
        case UP:
            return RIGHT;
        case DOWN:
            return LEFT;
        case LEFT:
            return UP;
        case RIGHT:
            return DOWN;
    }
}

/**
 * Rotates a relative direction based on the current rotation
 */
DIRECTION HPModel::getRotatedRelativeDirection(DIRECTION relative_direction) {
    int rotateAmount;

    switch (rotation) {
        case NO_ROTATION:
            rotateAmount = 0;
            break;
        case ROTATE_90:
            rotateAmount = 1;
            break;
        case ROTATE_180:
            rotateAmount = 2;
            break;
        case ROTATE_270:
            rotateAmount = 3;
            break;
    }

    DIRECTION new_relative_direction = relative_direction;
    for (int i = 0; i < rotateAmount; i++) {
        new_relative_direction = rotateRelativeDirection90Degrees(new_relative_direction);
    }

    return new_relative_direction;
}

std::string HPModel::pointsToJson() const {
    std::string json = "[\n";

    for (size_t i = 0; i < acids.size(); ++i) {
        const Acid &p = acids[i];

        json += "  {\n";
        json += "    \"x\": " + std::to_string(p.x) + ",\n";
        json += "    \"y\": " + std::to_string(p.y) + ",\n";
        json += "    \"z\": " + std::to_string(p.z) + ",\n";
        json += "    \"hydrophobic\": " + std::string(p.hydrophobic ? "true" : "false") + "\n";
        json += "  }";

        if (i < acids.size() - 1) {
            json += ",";
        }
        json += "\n";
    }

    json += "]";
    return json;
}

int HPModel::getScore() {
    return scoringPairs.size();
}

void HPModel::doFolds(const std::string &folds) {
    for (auto f : folds) {
        if (f == 'f') {
            fold(FORWARD);
        } else if (f == 'l') {
            fold(LEFT);
        } else if (f == 'r') {
            fold(RIGHT);
        } else if (f == 'u') {
            fold(UP);
        } else if (f == 'd') {
            fold(DOWN);
        } else {
            throw std::invalid_argument("Invalid fold direction, only folds allowed are 'flrud' (all lower case)");
        }
    }
}

// We need this for brute the brute force algorithm to allow multiple models to share earlier states

HPModel HPModel::clone() {
    HPModel newModel = HPModel(this->sequence);
    newModel.currentDirection = this->currentDirection;
    newModel.previousDirection = this->previousDirection;
    newModel.rotation = this->rotation;
    // Note that Acid *should* be trivially copyable as it contains only ints and bools
    newModel.acids = std::vector<Acid>(this->acids);
    newModel.acidMap = std::map<Point, Acid>(this->acidMap);
    newModel.scoringPairs = std::set<ScorePair>(this->scoringPairs);
    return newModel;
}

